package gemrot;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.JButton;

public class ListenerProblems {
	
	public static int eventConsumersAlive = 0;
	
	public static interface Listener {
		public void event();
	}
	
	public static class EventGenerator {
		
		List<Listener> strong = new ArrayList<Listener>();
		
		List<WeakReference<Listener>> weak = new ArrayList<WeakReference<Listener>>();
		
		public int weakSize() {
			return weak.size();
		}
		
		public int recountWeakListeners() {
			Iterator<WeakReference<Listener>> i = weak.iterator();
			int live = 0;
			while (i.hasNext()) {
				WeakReference<Listener> listener = i.next();
				if (listener == null) i.remove();
				else live += 1;
			}
			return live;
		}
		
		public void addWeakListener(Listener a) {
			this.weak.add(new WeakReference<Listener>(a));
		}
		
		public void addStrongListener(Listener a) {
			this.strong.add(a);
		}
		
		public int sizeStrong() {
			return this.strong.size();
		}
		
	}
	
	public static class EventConsumer implements Listener {
		
		public int weakAnonymousListenersAlive = 0;
		public int strongAnonymousListenersAlive = 0;
		
		public EventConsumer() {
			++ListenerProblems.eventConsumersAlive;
		}
		
		public void finalize() {
			--ListenerProblems.eventConsumersAlive;
		}
		
		public void registerWeakAnonymousListeners(int count, EventGenerator generator) {
			for (int i = 0; i < count; ++i) {
				generator.addWeakListener(new Listener() {
					@Override
					public void event() {
						System.out.println("anonymous listener, weak: event...");		
					}
					public void finalize() {
						--weakAnonymousListenersAlive;
					}
				});
				++weakAnonymousListenersAlive;
			}
		}
		
		public void registerStrongAnonymousListeners(int count, EventGenerator generator) {
			for (int i = 0; i < count; ++i) {
				generator.addStrongListener(new Listener() {
					@Override
					public void event() {
						System.out.println("anonymous listener, strong: event...");		
					}
					public void finalize() {
						--strongAnonymousListenersAlive;
					}
				});
				++strongAnonymousListenersAlive;
			}
		}

		@Override
		public void event() {
			System.out.println("event...");			
		}
		
	}
	
	public static void gc() {
		int count = 10;
		System.out.println(count + "x -> invoke gc(), sleep for 500ms.");
		for (int i = 0; i < count; i++) {
			System.gc();
			try {
				Thread.sleep(500);
			} catch (InterruptedException e) {
			}			
		}
	}
	
	public static void intro(){
		System.out.println("Introduction:");
		System.out.println("=============");
		System.out.println("Here we will be simulating problems with event listeners created as anonymous instances.");
		System.out.println("We will be using two terms ... weak-listener   -> added to the event generator through WeakReference.");
		System.out.println("                               strong-listener -> added to the event generator as a normal reference.");
		System.out.println();
	}
	
	public static void main(String[] args) {
		
		intro();
		
		System.out.println("Start of the simulation:");
		System.out.println("Consumers alive: " + eventConsumersAlive);
		
		EventGenerator generator = new EventGenerator();
		
		System.out.println("");
		System.out.println("1st scenerio:");
		System.out.println("-------------");
		System.out.println("Let's create a consumer and register strong-anonymous listener from within this consumer.");
		System.out.println("Then drop the reference on the consumer and let's see if it survives in the memory...");
		System.out.println("");
		
		
		System.out.println("Creating the consumer.");
		EventConsumer temp = new EventConsumer();
		System.out.println("Consumers alive: " + eventConsumersAlive);
		System.out.println("Registering strong-anonymous listener.");
		temp.registerStrongAnonymousListeners(1, generator);
		System.out.println("Dropping the reference.");
		temp = null;

		gc();

		System.out.println("Consumers alive: " + eventConsumersAlive);
		if (eventConsumersAlive > 0) {
			System.out.println("Damn ... the consumer has survived! We're leaking the memory here!");
		} else {
			System.out.println("Consumer has been gc()ed, how come???");
			System.exit(1);			
		}
		System.out.println();
		
		System.out.println("Let's drop the reference onto the event generator that is holding the reference on the listener that is keeping our consumer alive.");
		generator = null;
		
		gc();
		
		System.out.println("Consumers alive: " + eventConsumersAlive);		
		if (eventConsumersAlive == 0) {
			System.out.println("The consumer has ben gc()ed, good.");
			System.out.println("But hey... how often the event-generator is dropped in the real life, like instances of JButtons and such?");
		} else {
			System.out.println("Consumer hasn't been gc()ed, how come???");
			System.exit(1);			
		}
		
		
		generator = new EventGenerator();
		
		System.out.println("");
		System.out.println("2nd scenerio:");
		System.out.println("-------------");
		System.out.println("Let's create consumer and add weak-anonymous listener.");
		System.out.println("");
		
		
		System.out.println("Creating the consumer.");
		temp = new EventConsumer();
		System.out.println("Consumers alive: " + eventConsumersAlive);
		System.out.println("Registering weak-anonymous listener.");
		temp.registerWeakAnonymousListeners(1, generator);
		System.out.println("Dropping the reference.");
		temp = null;

		gc();

		System.out.println("Consumers alive: " + eventConsumersAlive);
		if (eventConsumersAlive == 0) {
			System.out.println("Consumer has been gc()ed, great!");
			System.out.println("");
			System.out.println("The problem of this solution is that if you're writing a library you can't implement");
			System.out.println("listeners list as a list of weak references, because you have to sustain event those anonymous (unreferenced) listeners.");
			System.out.println("Another bad thing is ... you can't register WeakReference<Listener> as a listener :-(");
			System.out.println("And that is really stupid ... because it is very elegant to have list of weak references to the listeners,");
			System.out.println("and let them be gc()ed, when they are not used anymore.");
			
		} else {
			System.out.println("Damn ... the consumer has survived, how come???");
			System.exit(1);			
		}
		System.out.println();
	}
	

}
